#!/usr/bin/perl
#
# Control the heater

# Strict and warnings are recommended.
use strict;
use warnings;
use File::ReadBackwards;
use Device::SerialPort qw( :PARAM :STAT 0.07 );
################################################################################
# Configuration for this script
# File name of the log of this script
my $thislog = '/opt/fhem/log/heatcontrol_perl.log';
# Log file for serial port
my $logfile_ser = "/opt/fhem/log/logfile_ser_perl.log";
# Log directory
my $logdir = "/opt/fhem/log/";
# Characteristic for the temperature setpoint
my $actmin = 5;
my $actmax = 100;
my $tmin = 35;
my $tmax = 60;
my $sumact;
################################################################################
# Given a path to the log directory and the names of the logged devices, create
# an array that contains the log filenames with their path
#
# Initialize array of log filenames
my @logfiles = ();
# Logged devices
# 1157: Living room
# 281a: Bedroom
# 5459: Attic
# 2346: Bathroom
my @devnames = ("1157", "281a", "5459", "2346");
# Loop for all log devices
for (my $i = 0; $i < @devnames; $i++) {
  # Actual device name
  my $devname = $devnames[$i];
  # Get actual year
  my @dt = localtime();
  my $cyear = $dt[5] + 1900;
  #print $cyear, "\n";
  # Make a string with the actual file name
  my $tmp = sprintf('%sFHT_%s-%s.log', $logdir, $devname, $cyear);
  # Add file name to the array
  push @logfiles, $tmp;

  print @logfiles, "\n";
}

################################################################################
# Main cycle
while (1) {
  # Open log file in append mode
  open(my $fh, '>>', $thislog) or die "Could not open file '$thislog' $!";
  # Open serial port log file in append mode
  open(my $fhs, '>>', $logfile_ser) or die "Could not open file '$logfile_ser' $!"; 
  ################################################################################
  # Given a an array of log filenames, extracts the last line that contains the
  # keyword "actuator:" from each log file and from this line reads the value
  # given after the keyword. Remove the units and store this value in a new array
  #
  # Loop for all log files
  # Initialize actuator values array
  my @acts = ();
  # Initialize actual actuator value
  my $act = 0;
  for (my $i = 0; $i < @logfiles; $i++) {
    # Log filename
    my $logfile = $logfiles[$i];
      # Open log file in backwards mode
    my $bw = File::ReadBackwards->new($logfile)
             or die "can't read 'log_file' $!";
    # Initialize counter for while loop
    my $i = 0;
    # Maximum number of allowed iterations
    my $nmax = 10;
    # Read lines of log file in backwards mode, as long as there are lines to read
    # or the counter did not reach its maximum value
    while( (defined(my $log_line = $bw->readline) && ($i < $nmax) ) ) {
      print $i, ' ', $log_line;
      # Check if this line contains the keyword "actuator:" and not the error message "unknown"
      if (($log_line =~ /actuator:/) && !($log_line =~ /unknown/)) {
        print $log_line;
        # Split the line in words ...
        my @words = split / /, $log_line;
        # ... and take the last one
        my $act_with_units = $words[-1];
        # Remove unit value ("%")
        my @actword = split /%/, $act_with_units;
        $act = $actword[0];
        print $act, "\n";
        # Set counter to the max value to exit from the while loop
        $i = $nmax;
      }
      # Increase the counter
      $i = $i + 1;
    }
    # Write value to an array
    push @acts, $act;
  }

  ##############################################################################
  # Calculate the temperature setpoint for the boiler
  # Apply gains depending on the room: living room 1, the others 1/3
  # Loop for all the actuator values
  my @actsw = ();
  my $actw;
  for (my $i = 0; $i < @acts; $i++) {
    if ($i == 0) {
      $actw = $acts[$i];
    } else {
      $actw = $acts[$i] / 3;
    }
  push @actsw, $actw;
  }
  # Calculate the sum of all the actuator values ($sumact)
  $sumact = 0;
  # Loop for all the actuator values
  for (my $i = 0; $i < @actsw; $i++) {
    $sumact = $sumact + $actsw[$i];
  }

  print "Debug start sumact\n";
  print $sumact, "\n"; 
  print "Debug finish sumact\n"; 

  # Interpolate to calculate the temperature setpoint ($ttar)
  my $ttar = 0;
  # If the sum is larger then the max value use the max value ...
  if ($sumact > $actmax) {
    $ttar = $tmax;
  } else {
    # ... otherwise check if larger than the min value ...
    if ($sumact >= $actmin) {
    # ... and in this case interpolate
    $ttar = $tmin + ($sumact - $actmin) * ($tmax - $tmin) / ($actmax - $actmin);
    }
  }
  # Convert to string that contains the integer value
  $ttar = sprintf('%02.0f', $ttar);

  print "Debug start ttar\n";
  print $ttar, "\n";
  print "Debug finish ttar\n";

  
  ################################################################################
  # Store actuator values in a log file
  #


  my @dt = localtime();
  my $hou = sprintf('%02.0f', $dt[2]);
  my $min = sprintf('%02.0f', $dt[1]);
  my $sec = sprintf('%02.0f', $dt[0]);
  print $fh "$hou:$min:$sec ";

  #print $fh  "Before write log: $hou:$min:$sec \n";

  print $fh "Actuators:";

  # Loop for all the actuator values
  for (my $i = 0; $i < @acts; $i++) {
    # Put value
    print $fh sprintf(' %3.0f', $acts[$i]);
  }
  # Log charachteristic values actuators sum and target temperature
  print $fh sprintf(' actmin: %3.0f', $actmin);
  print $fh sprintf(' actmax: %3.0f', $actmax);
  print $fh sprintf(' tmin: %02.0f', $tmin);
  print $fh sprintf(' tmin: %02.0f', $tmax);
  print $fh sprintf(' sumact:  %3.0f', $sumact);
  print $fh sprintf(' ttar: %02.0f', $ttar);  

  # Add an end of line
  print $fh "\n";

  #print $fh  "After write log: $hou:$min:$sec \n";

  ###############################################################################
  # Initialize and open serial port
  # Initialize port
  my $port;
  $port = Device::SerialPort ->new("/dev/ttyUSB0") || die "Can't open $port: $!";
  my $portbr = $port->baudrate(9600);
  my $portpa = $port->parity("none");
  my $portdb = $port->databits(8);
  my $portsb = $port->stopbits(1);
  my $porths = $port->handshake("rts");
  my $portrc = $port->read_char_time(1000);
  $port->write_settings() || die("settings failed\n");;
  # Log port configuration
  # Write log
  print $fhs "Port open - baud: $portbr parity: $portpa databits: $portdb stopbits: $portsb handshake: $porths \n";
  # Give time to the serial port to become ready
  sleep(1);
  
  # Initialize variables
  my $ttarstr = sprintf("%02.0f", $ttar);
  my $cmdstr;
  my $ReadLines;
  my $BytePerLine = 11;
  my $i = 0;
  my $imax = 4;

  # Put time on log file
  print $fhs "$hou:$min:$sec\n";

  my $InBytes;
  my $string_in;

  # Do $imax attempts to send command to the gateway
  while ($i < $imax) {
    # Read one line from serial port to syncronize
    $ReadLines = 1;
    $InBytes = $ReadLines * $BytePerLine;
    $string_in = $port->read($InBytes) || die "Can't read from port: $!" ;
    print $fhs $string_in, "\n";
    # Wait 100ms
    sleep(0.1);
    # Make command string
    $cmdstr = "CS=$ttarstr\r\n";
    print "$cmdstr\n";
    # Send command
    $port->write($cmdstr);
    # Read next 3 lines
    $ReadLines = 3;
    $InBytes = $ReadLines * $BytePerLine;
    $string_in = $port->read($InBytes) || die "Can't read from port: $!" ;
    print $fhs $string_in, "\n";
    # Check if there was a positive response from the gateway
    if ($string_in =~ /CS:/) {
      # In this case log it ...
      print $fhs "Attempt $i: command successfull\n";
      # ... and set counter to exit from this loop
      $i = $imax;
    }
    else {
      # In this case increase counter
      $i = $i + 1;
      # If the counter reached the $imax ...
      if ($i == $imax) {
        # ... log it
        print $fhs "Failed writing command after $i attempts\n";
      }
    }
  }

  # The following is to read values from the boiler but it is not ready yet.
  # Now it is deactivated, to reactivate set $imax = 20 (or a number > 0)
  $i = 0;
  $imax = 0;
  my $cvtemp;
  while ($i < $imax) {
    sleep(0.1);
    $ReadLines = 1;
    $InBytes = $ReadLines * $BytePerLine;
    $string_in = $port->read($InBytes) || die "Can't read from port: $!" ;
    print $string_in, "\n";
    if ($string_in =~ /(B)(.)(019)(.{2})(.{2})/) {
      #my $cvtemphex = $4;
      #
      #print "$4\n";
      my $hb = $4;
      my $lb = $5;

      print "$i $hb$lb\n";

      $cvtemp = hex("0x$hb") + hex("0x$lb")/255;
      $i = $imax;
    } else {
      $cvtemp = "0";
      $i = $i + 1;
    }
  }


  #print "Measured cd temp: $cvtemp\n";


  ##############################################################################
  # Before exiting the main cycle
  # Close this log file
  close($fh);
  # Close serial port log file
  close($fhs);
  # Close serial port
  $port->close;

  my @dtb = localtime();
  my $houb = sprintf('%02.0f', $dtb[2]);
  my $minb = sprintf('%02.0f', $dtb[1]);
  my $secb = sprintf('%02.0f', $dtb[0]);
  print "localtime before sleep:  $houb:$minb:$secb\n";
 
  #print "localtime before sleep: ";
  #print localtime(), "\n";

  # Wait 1 minute
  sleep (60);

  my @dta = localtime();
  my $houa = sprintf('%02.0f', $dta[2]);
  my $mina = sprintf('%02.0f', $dta[1]);
  my $seca = sprintf('%02.0f', $dta[0]);
  print "localtime after sleep:  $houa:$mina:$seca\n";


  #print "localtime after sleep: "
  #print localtime(), "\n";

}
################################################################################
# Before closing the script
#
